import java.awt.*;

import scheduler.*;
import sysPr.*;

public class SystemThread extends SimProcess {
  int n = ((Algorithm)(scheduler.system)).getNumberOfNodes();
  int c;
  int round = 0;
  CommunicationGraph communicationGraph;
  int initiator = 0;
  int low = 3, high = 5;
  boolean[][] received;
  boolean[][] sent;
  int[][][] failure = {{{}, {}, {}, {}, {7}, {7}, {7}}  // round 1
/*                       , {{0, 1, 2, 3, 4, 5, 6, 7},        // round 2
                        {0, 1, 2, 3, 4, 5, 6, 7},
                        {0, 1, 2, 3, 4, 5, 6, 7},
                        {0, 1, 2, 3, 4, 5, 6, 7},
                        {0, 1, 2, 3, 4, 5, 6, 7},
                        {0, 1, 2, 3, 4, 5, 6, 7},
                        {0, 1, 2, 3, 4, 5, 6, 7}},
*/                      };

  public SystemThread(String name, Scheduler scheduler, int nodeNumber, CommunicationGraph communicationGraph) {
    super(name, scheduler, nodeNumber);
    ((Algorithm)(scheduler.system)).addRow(name, nodeNumber, false);
    scheduler.createThread(this);
    this.communicationGraph = communicationGraph;
  }

  public void runProcess() {
    int i;
    Connection connection;
    boolean tmp, committed = false;
    received = new boolean[n][n + 1];
    sent = new boolean[n][n + 1];
    for  (i = 0; i < n; i++) {
      for (int j = 0; j <= n; j++) {
        received[i][j] = false;
        sent[i][j] = false;
      }
    }
    while (scheduler.system.getSimulationTime() < 1) scheduler.schedule();

    while (true) {
      // SystemPrinter.println("Waiting for Message");
      ((Algorithm)(scheduler.system)).getGraphFrame().stateChanged();
      Message message = receive();
      int sender = message.source.nodeNumber;

      if (message.type == "NEXTROUND") {
        round++;
        int f = 1;
        if (round > 2 * f + 3) {
          // if not already decided then decide for zero'
          if (commits()) {
            ((Algorithm)(scheduler.system)).getGraphFrame().mysim.ms.myNodes[nodeNumber].set_color(Color.red);
          } else {
            ((Algorithm)(scheduler.system)).getGraphFrame().mysim.ms.myNodes[nodeNumber].set_color(Color.green);
          }
          return;
        }
        if (nodeNumber == 0 && failure.length > (round - 1)) {
          SystemPrinter.println("Round " + (round - 1) + ", " + failure[round - 1].length);
          if (failure[round - 1].length > 0) {
            for (i = 0; i < failure[round - 1].length; i++) {
              if (failure[round - 1][i].length > 0) {
                SystemPrinter.println("Round " + (round - 1) + ", destination = " + i + ", " + failure[round - 1][i].length);
                for (int j = 0; j < failure[round - 1][i].length; j++) {
                  SystemPrinter.println("Round " + (round - 1) + ", destination = " + i
                     + ", message = " + failure[round - 1][i][j]);

                  if (!sent[i][failure[round - 1][i][j]]) {
                    for (int k = 0; k < outgoingConnections.size(); k++) {
                      if ((connection = (Connection)(outgoingConnections.elementAt(k)))
                            .destination.nodeNumber == i) {
                        if (failure[round - 1][i][j] == n) {
                          connection.send("STAR");
                          sent[connection.destination.nodeNumber][n] = true;
                        } else {
                          connection.send("J", failure[round - 1][i][j]);
                          sent[connection.destination.nodeNumber][failure[round - 1][i][j]] = true;
                        }
                      }
                    }
                  }
                }
              }
            }
          }
        }
        // M1
        if (initiates()) {
          for (i = 0; i < n; i++) {
            connection = (Connection)(outgoingConnections.elementAt(i));
            if (!sent[connection.destination.nodeNumber][n]) {
              connection.send("STAR");
              sent[connection.destination.nodeNumber][n] = true;
            }
          }
        }
        tmp = committed;
        committed = commits();
        if (!tmp && committed) {
          SystemPrinter.println("******************** committed");
          ((Algorithm)(scheduler.system)).getGraphFrame().mysim.ms.myNodes[nodeNumber].set_color(Color.red);
        }

        // M2: direkte Untersttzung
        boolean[] m = witnesses(n);
        SystemPrinter.print("Direkt untersttzt: ", true);
        for (int j = 0; j < n; j++) {
          if (m[j]) {
            for (i = 0; i < n; i++) {
              connection = (Connection)(outgoingConnections.elementAt(i));
              if (!sent[connection.destination.nodeNumber][j]) {
                connection.send("J", j);
                sent[connection.destination.nodeNumber][j] = true;
              }
            }
            SystemPrinter.print((j + 1) + " ", false);
          }
        }
        SystemPrinter.print("\n", false);

        // M3: indirekte Untersttzung
        for (int j = 0; j < n; j++) {
          if (card(witnesses(j)) >= low) {
            for (i = 0; i < n; i++) {
              connection = (Connection)(outgoingConnections.elementAt(i));
              if (!sent[connection.destination.nodeNumber][j]) {
                connection.send("J", j);
                sent[connection.destination.nodeNumber][j] = true;
              }
            }
          }
        }
      } else if (message.type == "STAR") {
        received[message.source.nodeNumber][n] = true;
      } else if (message.type == "J") {
        received[message.source.nodeNumber][message.winner] = true;
      } else {
	SystemPrinter.println("Unknown message type; suspending myself");
	scheduler.suspend();
      }
    }
  }
  // Compute set of nodes which sent x to this node
  public boolean[] witnesses(int x) {
    int i;
    boolean[] result = new boolean[n];
    for (i = 0; i < n; i++) {
      result[i] = received[i][x];
    }
    return result;
  }

  public boolean[] confirmed() {
    boolean[] result = new boolean[n];
    result[initiator] = false;
    for (int i = 0; i < n; i++) {
      if (i != initiator)
	result[i] = (card(witnesses(i)) >= high);
    }
    return result;
  }

  public int card(boolean[] set) {
    int result = 0;
    for (int i = 0; i < n; i++) {
      if (set[i]) result++;
    }
    return result;
  }

  public boolean initiates() {
    boolean[] w;
    w = witnesses(n);
    if (w[nodeNumber]) {               // initiated by (I1)
      SystemPrinter.println("initiated  by I1");
      return true;
    }

    if (card(confirmed()) >= low + ((round + 1) / 2 >= 2 ? (round + 1) / 2 - 2 : 0)) {       // initiated by (I2)
      SystemPrinter.println("initiated  by I2");
      return true;
    }
    if (round == 2 && w[initiator]) {  // initiated by (I3)
      SystemPrinter.println("initiated  by I3");
      return true;
    }
    return false;
  }

  public boolean commits() {
    int result = 0;
    for (int i = 0; i < n; i++) {
      if (card(witnesses(i)) >= high) result++;
    }
    return (result >= high ? true : false);
  }

  public void printreceived() {
    for (int j = 0; j < n; j++) {
      SystemPrinter.print("received from " + j + ": " + received[j][0], true);
      for (int i = 1; i <= n; i++) SystemPrinter.print(", " + received[j][i], false);
      SystemPrinter.print("\n", false);
    }
  }
}
