package scheduler;

import java.util.*;
import java.awt.*;
import sysPr.*;

import java.applet.*;

public class Scheduler {
  static final int RUNNING     = 1;
  static final int READY       = 2;
  static final int BLOCKED     = 3;
  static final int TIMED       = 4;
  static final int NONEXISTENT = 5;
  static final int NUMBEROFTHREADS = 100;
  public static final int SIGNAL = 1;
  public static final int TIMEOUT = 2;
  private Date now;
  private int turn = -1;
  public Application system;
  private Thread threads[] = new Thread[NUMBEROFTHREADS];
  private String names[] = new String[NUMBEROFTHREADS];
  private int states[] = new int[NUMBEROFTHREADS];
  private int resumes[] = new int[NUMBEROFTHREADS];
  private int resumedBy[] = new int[NUMBEROFTHREADS];
  private Date dates[] = new Date[NUMBEROFTHREADS];
  private int interruptible[] = new int[NUMBEROFTHREADS];
  private int threadCount = 0;
  public boolean stopped = true;
  public int deltaT;
  Date stoppingTime, continueingTime;
  
  private int pid(String name) {
    for (int i = 0; i < threadCount; i++) {
//       System.out.println("pid: " + i + "  " + name +  "  " + names[i]);
      if (name.equals(names[i])) return i;
    }
    return -1;
  }

  public Scheduler(Application cs, int numberOfShownProcesses) {
    system = cs;
    for (int i = 0; i < 100; i++) {
      resumes[i] = 0;
      dates[i] = new Date();
      interruptible[i] = 0;
//      if (i < numberOfShownProcesses) {
//        system.updateState(i, names[i], Color.white,
//                           "                ");
//      }
    }
  }

  public synchronized void schedule() {
    Thread caller;
    int callerPid;
    int i, n;
    boolean found = false;
    caller = Thread.currentThread();
    callerPid = pid(caller.getName());
// SystemPrinter.println("name = " + caller.getName() + "; callerPid = " + callerPid
// 		      + "; turn = " + turn);
    if (callerPid < 0) {
      if (turn >= 0) return;
    }
    if (stopped) {
// SystemPrinter.println("Stopped == true; callerPID = " + callerPid);
      if (callerPid >= 0) {
        while (states[callerPid] != RUNNING) {
          try {
            wait();
          } catch (InterruptedException e) { }
        }
      }
    } else {
// SystemPrinter.println("Stopped == false");
      for (i = 0; i < threadCount; i++) {
        if (states[i] == RUNNING) {
          if (interruptible[i] > 0) return;
          states[i] = READY;
          system.updateState(i, names[i], Color.orange,
                             "                ");
        }
      }

// Resume processes with elapsed waiting time
      now = new Date();
      for (i = 0; i < threadCount; i++) {
        if (now.after(dates[i]) && states[i] == TIMED) {
          states[i] = READY;
	  resumedBy[i] = TIMEOUT;
          system.updateState(i, names[i], Color.orange, "                ");
        }
      }
  
// SystemPrinter.println("Select process to be run");
      found = false;
      for (i = 1; i <= threadCount; i++) {
        n = (turn + i) % threadCount;
        if (threads[n] != null && states[n] == READY) {
          turn = n;
          found = true;
          break;
        }
      }
    }
    if (!found) {
      turn = -1;
      if (callerPid < 0) return;
    }

// Resume elected process
    if (found) {
      states[turn] = RUNNING;
      system.updateState(turn, names[turn], Color.green,
                         "                ");
      notifyAll();
    }
    if (callerPid >= 0) {
      while (states[callerPid] != RUNNING) {
        try {
          wait();
        } catch (InterruptedException e) { }
      }
    }
//   Wait some time after scheduling action
    double x = 0;
    for (int j = 0; j < 100000; j++) {
      x = (x + 2.0) / 3.0;
    }
    return;
  }
  
  public synchronized void createThread(Thread p) {
    if (threadCount == NUMBEROFTHREADS) {
      SystemPrinter.println("Too much Threads");
      System.exit(1);
    }
    threads[threadCount] = p;
    names[threadCount] = p.getName();
    states[threadCount] = READY;
//    system.updateState(threadCount, names[threadCount], Color.orange,
//                       "                ");
    threadCount++;
    p.start();
  }

  public synchronized void eliminateThread(Thread p)
  { int n = 0, i = 0;
    threads[i = pid(p.getName())] = null;
    states[i] = NONEXISTENT;
    system.updateState(i, names[i], Color.white,
                       "                ");
    while (threads[n] == null) {
      n++;
      if (n >= threadCount) {
        while (true) ;
      }
    }
    do {
      turn = (turn + 1) % threadCount;
    } while (threads[turn] == null);
    notifyAll();
  }

  public synchronized void initialize() {
    stopped = false;
    turn = -1;
    // SystemPrinter.println("notify all processes");
    notifyAll();
  }

  public synchronized void suspend() {
    int callerPid;
    callerPid = pid(Thread.currentThread().getName());
    resumes[callerPid]--;
    if (resumes[callerPid] < 0) {
      states[callerPid] = BLOCKED;
      interruptible[callerPid] = 0;
      system.updateState(callerPid, names[callerPid], Color.red, "resume");
      system.updateButton(callerPid, "resume", true, Color.green);
    }
    schedule();
  }

  public synchronized void resume(String name) {
    int i = pid(name);
    resumes[i]++;
    states[i] = READY;
    resumedBy[i] = SIGNAL;
    system.updateState(i, names[i], Color.orange,
                       "                ");
    system.updateButton(i, "                ", false, Color.gray);
    schedule();
  }

  public synchronized void sleep(int msec)
  { int callerPid;
    callerPid = pid(Thread.currentThread().getName());
    Date now = new Date();
    dates[callerPid] = new Date(now.getTime() + msec);
    states[callerPid] = TIMED;
    system.updateState(callerPid, names[callerPid], Color.magenta,
                       "                ");
    schedule();
  }
  
  public synchronized int continuedBy()
  { int callerPid;
    callerPid = pid(Thread.currentThread().getName());
    return resumedBy[callerPid];
  }
  
  public synchronized void stop() {
    Date now = new Date();  
    stopped = true;
    stoppingTime = new Date(now.getTime());
    // SystemPrinter.println("Stopped at time " + stoppingTime.toString());
  }
  
  public synchronized void goOn() {
    Date now = new Date();
    continueingTime = now;
    stopped = false;
    // SystemPrinter.println("Continued at time " + now.toString());
    long delta = continueingTime.getTime() - stoppingTime.getTime();
    // SystemPrinter.println("System was stopped for " + delta + " msec");
    for (int i = 0; i < threadCount; i++) {
      if (now.after(dates[i]) && states[i] == TIMED) {
        dates[i] = new Date(now.getTime() + delta);
      }
    }
  }
  
  public synchronized void changeTimeScale(int delta) {
    deltaT = delta;
    // SystemPrinter.println("deltaT = " + deltaT);
  }
  
  public int myPid() {
    return pid(Thread.currentThread().getName());
  }
  
  public void beginOfAtomic() {
    interruptible[myPid()]++;
  }
  public void endOfAtomic() {
    interruptible[myPid()]--;
  }
}
